/*
  $Header: /home/richard/myntp/chrony/chrony-1.0/RCS/manual.c,v 1.12 1998/05/20 06:10:52 richard Exp $

  =======================================================================

  chronyd/chronyc - Programs for keeping computer clocks accurate.

  Copyright (C) 1997-1998 Richard P. Curnow <richard@curnow.demon.co.uk>
  All rights reserved.

  For conditions of use, refer to the file LICENCE.

  =======================================================================

  Routines for implementing manual input of real time.

  The daemon accepts manual time input over the control connection,
  and adjusts the system time to match.  Besides this, though, it can
  determine the average rate of time loss or gain of the local system
  and adjust the frequency accordingly.

  */

#include <stddef.h>

#include "manual.h"
#include "logging.h"
#include "local.h"
#include "conf.h"
#include "util.h"
#include "ntp.h"
#include "reference.h"

static int enabled = 0;

/* More recent samples at highest indices */
typedef struct {
  struct timeval when; /* This is our 'cooked' time */
  double offset; /* + if we are fast of the supplied reference */
} Sample;

#define MAX_SAMPLES 16

static Sample samples[16];
static int n_samples;

static int replace_margin;
static int error;

/* Eventually these constants need to be user-defined in conf file */
#define REPLACE_MARGIN 300
#define ERROR 0.2

/* ================================================== */

static void
slew_samples(struct timeval *raw,
             struct timeval *cooked, 
             double dfreq,
             double afreq,
             double doffset,
             int is_step_change,
             void *not_used);

/* ================================================== */

void
MNL_Initialise(void)
{
  if (CNF_GetManualEnabled()) {
    enabled = 1;
  } else {
    enabled = 0;
  }

  n_samples = 0;

  replace_margin = REPLACE_MARGIN;
  error = ERROR;

  LCL_AddParameterChangeHandler(slew_samples, NULL);

  return;
}

/* ================================================== */

void
MNL_Finalise(void)
{
  return;
}

/* ================================================== */

int
MNL_AcceptTimestamp(struct timeval *ts, long *offset_cs)
{
  struct timeval now;
  double local_clock_err;
  int diff_secs;
  double offset;
  int replace_sample;
  int i, n;
  double doffset, dt, freq = 0.0;
  double skew = 999.9;
  int found_freq;
  double margin;

  if (enabled) {

    /* Check whether timestamp is within margin of old one */
    LCL_ReadCookedTime(&now, &local_clock_err);

    UTI_DiffTimevalsToDouble(&offset, &now, ts);

    if (n_samples > 0) {
      diff_secs = now.tv_sec - samples[n_samples - 1].when.tv_sec;
      if ((diff_secs >= -replace_margin) &&
          (diff_secs <= replace_margin)) {
        replace_sample = 1;
      } else {
        replace_sample = 0;
      }
    } else {
      replace_sample = 0;
    }

    if (replace_sample) {
      samples[n_samples - 1].when = now;
      samples[n_samples - 1].offset = offset;
    } else {

      /* Check if buffer full up */
      if (n_samples == MAX_SAMPLES) {
        /* Shift samples down */
        for (i=1; i<n_samples; i++) {
          samples[i-1] = samples[i];
        }
        --n_samples;
      }

      samples[n_samples].when = now;
      samples[n_samples].offset = offset;
      ++n_samples;
    }

    /* Do scan through data to look for whether any frequency estimate
       is significantly non-zero */
    found_freq = 0;
    n = n_samples - 1;
    for (i=0; i<n; i++) {
      UTI_DiffTimevalsToDouble(&dt, &samples[n].when, &samples[i].when);
      doffset = samples[n].offset - samples[i].offset;
      if (dt >= 0.0) {
        margin = 2.0 * error / dt;
        freq = doffset / dt;
        skew = 2.0 * error / dt;
        LOG(LOGS_INFO, LOGF_Manual, "Between %s and %s freq=%.6fppm\n",
            UTI_TimevalToString(&samples[i].when), 
            UTI_TimevalToString(&samples[n].when),
            1.0e6 * freq);
        if (((freq - margin) > 0.0) ||
            ((freq + margin) < 0.0)) {
          found_freq = 1;
          break;
        }
      }
    }

    if (found_freq) {
      LOG(LOGS_INFO, LOGF_Manual, "Making a frequency change of %.3fppm and a slew of %.6f\n", 1.0e6 * freq, offset);
      REF_SetManualReference(&now,
                       offset,
                       freq, skew);
    } else {
      LOG(LOGS_INFO, LOGF_Manual, "Making a slew of %.6f\n", offset);
      REF_SetManualReference(&now,
                       offset,
                       0.0, skew);
    }

    *offset_cs = (long)(0.5 + 100.0 * offset);

    return 1;

  } else {
  
    return 0; /* until we implement it */

  }
}

/* ================================================== */

static void
slew_samples(struct timeval *raw,
             struct timeval *cooked, 
             double dfreq,
             double afreq,
             double doffset,
             int is_step_change,
             void *not_used)
{
  double elapsed, delta_time;
  int i;
  for (i=0; i<n_samples; i++) {
    UTI_DiffTimevalsToDouble(&elapsed, cooked, &samples[i].when);
    delta_time = elapsed * dfreq - doffset;
    UTI_AddDoubleToTimeval(&samples[i].when, delta_time, &samples[i].when);
    samples[i].offset += delta_time;
  }
  return;
}

/* ================================================== */

void
MNL_Enable(void)
{
  enabled = 1;
}


/* ================================================== */

void
MNL_Disable(void)
{
  enabled = 0;
}

/* ================================================== */

void
MNL_Reset(void)
{
  n_samples = 0;
}

