/*
  $Header: /home/richard/myntp/chrony/chrony-1.0/RCS/rtc_linux.c,v 1.8 1998/05/29 05:15:32 richard Exp $

  =======================================================================

  chronyd/chronyc - Programs for keeping computer clocks accurate.

  Copyright (C) 1997-1998 Richard P. Curnow <richard@curnow.demon.co.uk>
  All rights reserved.

  For conditions of use, refer to the file LICENCE.

  =======================================================================

  Real-time clock driver for linux.  This interfaces the program with
  the clock that keeps time when the machine is turned off.

  */

#include "options.h"

#if defined LINUX

#include <stdio.h>
#include <math.h>
#include <sys/time.h>
#include <sys/types.h>
#include <linux/mc146818rtc.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>
#include <string.h>

#include "logging.h"
#include "sched.h"
#include "local.h"
#include "util.h"
#include "sys_linux.h"
#include "regress.h"
#include "rtc.h"
#include "rtc_linux.h"
#include "conf.h"
#include "mkdirpp.h"

/* ================================================== */
/* Forward prototypes */

static void measurement_timeout(void *any);

static void read_from_device(void *any);

/* ================================================== */

typedef enum {
  OM_NORMAL,
  OM_INITIAL,
  OM_AFTERTRIM
} OperatingMode;

static OperatingMode operating_mode = OM_NORMAL;

/* ================================================== */

static int fd = -1;

#define LOWEST_MEASUREMENT_PERIOD 15
#define HIGHEST_MEASUREMENT_PERIOD 480

/* Try to avoid doing regression after _every_ sample we accumulate */
#define N_SAMPLES_PER_REGRESSION 4

static int measurement_period = LOWEST_MEASUREMENT_PERIOD;

static int timeout_running = 0;
static SCH_TimeoutID timeout_id;

/* ================================================== */

/* Maximum number of samples held */
#define MAX_SAMPLES 64

/* Real time clock samples.  We store the seconds count as originally
   measured, together with a 'trim' that compensates these values for
   any steps made to the RTC to bring it back into line
   occasionally.  The trim is in seconds. */
static time_t rtc_sec[MAX_SAMPLES];
static double rtc_trim[MAX_SAMPLES];

/* Reference time, against which delta times on the RTC scale are measured */
static time_t rtc_ref;


/* System clock (gettimeofday) samples associated with the above
   samples. */
static struct timeval system_times[MAX_SAMPLES];

/* Number of samples currently stored. */
static int n_samples;   

/* Number of new samples since last regression */
static int n_samples_since_regression;

/* Number of runs of residuals in last regression (for logging) */
static int n_runs;

/* Coefficients */
/* Whether they are valid */
static int coefs_valid;

/* Reference time */
static time_t coef_ref_time;
/* Number of seconds by which RTC was fast of the system time at coef_ref_time */
static double coef_seconds_fast;

/* Estimated number of seconds that RTC gains relative to system time
   for each second of ITS OWN time */
static double coef_gain_rate;

/* Gain rate saved just before we step the RTC to correct it to the
   nearest second, so that we can write a useful set of coefs to the
   RTC data file once we have reacquired its offset after the step */
static double saved_coef_gain_rate;

/* Filename supplied by config file where RTC coefficients are
   stored. */
static char *coefs_file_name;

/* ================================================== */

static FILE *logfile=NULL;

/* ================================================== */

static void (*after_init_hook)(void *) = NULL;
static void *after_init_hook_arg = NULL;

/* ================================================== */

static void
discard_samples(int new_first)
{
  int n_to_save;

  assert(new_first < n_samples);
  assert(new_first >= 0);

  n_to_save = n_samples - new_first;

  memmove(rtc_sec, rtc_sec + new_first, n_to_save * sizeof(time_t));
  memmove(rtc_trim, rtc_trim + new_first, n_to_save * sizeof(double));
  memmove(system_times, system_times + new_first, n_to_save * sizeof(struct timeval));

  n_samples = n_to_save;
  return;
}

/* ================================================== */

#define NEW_FIRST_WHEN_FULL 4

static void
accumulate_sample(time_t rtc, struct timeval *sys)
{

  if (n_samples == MAX_SAMPLES) {
    /* Discard oldest samples */
    discard_samples(NEW_FIRST_WHEN_FULL);
  }

  rtc_sec[n_samples] = rtc;

  /* Always use most recent sample as reference */
  rtc_ref = rtc;

  rtc_trim[n_samples] = 0.0;
  system_times[n_samples] = *sys;
  ++n_samples;
  ++n_samples_since_regression;
  return;
  
}

/* ================================================== */
/* The new_sample flag is to indicate whether to adjust the
   measurement period depending on the behaviour of the standard
   deviation. */

static void
run_regression(int new_sample,
               int *valid,
               time_t *ref,
               double *fast,
               double *slope)
{
  double rtc_rel[MAX_SAMPLES]; /* Relative times on RTC axis */
  double offsets[MAX_SAMPLES]; /* How much the RTC is fast of the system clock */
  int i, n;
  double est_intercept, est_slope;
  int best_new_start;

  if (n_samples > 0) {

    n = n_samples - 1;

    for (i=0; i<n_samples; i++) {
      rtc_rel[i] = rtc_trim[i] + (double)(rtc_sec[i] - rtc_ref);
      offsets[i] = ((double) (rtc_ref - system_times[i].tv_sec) -
                    (1.0e-6 * (double) system_times[i].tv_usec) +
                    rtc_rel[i]);

    }

    if (RGR_FindBestRobustRegression
        (rtc_rel, offsets,
         n_samples, 1.0e-9,
         &est_intercept, &est_slope,
         &n_runs,
         &best_new_start)) {

      /* Calculate and store coefficients.  We don't do any error
         bounds processing on any of these. */
      *valid = 1;
      *ref = rtc_ref;
      *fast = est_intercept;
      *slope = est_slope;

      if (best_new_start > 0) {
        discard_samples(best_new_start);
      }


    } else {
      /* Keep existing coefficients. */
    }
  } else {
    /* Keep existing coefficients. */
  }

}

/* ================================================== */

static void
slew_samples
(struct timeval *raw, struct timeval *cooked,
 double dfreq, double afreq_ppm,
 double doffset, int is_step_change,
 void *anything)
{
  int i;
  double elapsed;
  double new_freq;
  double old_freq;
  double delta_time;
  double old_seconds_fast, old_gain_rate;

  new_freq = 1.0e-6 * afreq_ppm;
  old_freq = (new_freq - dfreq) / (1.0 - dfreq);

  for (i=0; i<n_samples; i++) {
    UTI_DiffTimevalsToDouble(&elapsed, cooked, system_times + i);

    delta_time = elapsed * dfreq - doffset;

    UTI_AddDoubleToTimeval(system_times + i, delta_time, system_times + i);

  }

  old_seconds_fast = coef_seconds_fast;
  old_gain_rate = coef_gain_rate;

  if (coefs_valid) {
    coef_seconds_fast += doffset;
    coef_gain_rate = 1.0 - ((1.0 + new_freq) / (1.0 + old_freq)) * (1.0 - coef_gain_rate);
  }

#if 0
  LOG(LOGS_INFO, LOGF_RtcLinux, "dfreq=%.8f doffset=%.6f new_freq=%.3f old_freq=%.3f old_fast=%.6f old_rate=%.3f new_fast=%.6f new_rate=%.3f\n",
      dfreq, doffset, 1.0e6*new_freq, 1.0e6*old_freq,
      old_seconds_fast, 1.0e6 * old_gain_rate,
      coef_seconds_fast, 1.0e6 * coef_gain_rate);
#endif

}

/* ================================================== */
/* file_name is the name of the file where we save the RTC params
   between executions.  Return status is whether we could initialise
   on this version of the system. */

int
RTC_Linux_Initialise(char *file_name)
{

  int major, minor, patch;
  char *direc;
  char filename[1024];

  /* Check whether we can support the real time clock.

     Linux 1.2.x - haven't checked yet

     Linux 1.3.x - don't know, haven't got a system to look at

     Linux 2.0.x - For x<=31, using any variant of the adjtimex() call
     sets the kernel into a mode where the RTC was updated every 11
     minutes.  The only way to escape this is to use settimeofday().
     Since we need to have sole control over the RTC to be able to
     measure its drift rate, and there is no 'notify' callback to warn
     you that the kernel is going to do this, I can't see a way to
     support this.

     Linux 2.0.x - For x>=32 the adjtimex()/RTC behaviour was
     modified, so that as long as the STA_UNSYNC flag is set the RTC
     is left alone.  This is the mode we exploit here, so that the RTC
     continues to go its own sweet way, unless we make updates to it
     from this module.

     Linux 2.1.x - don't know, haven't got a system to look at.

     */

  SYS_Linux_GetKernelVersion(&major, &minor, &patch);

  /* Obviously this test can get more elaborate when we know about
     more system types. */
  if ((major != 2) ||
      (minor != 0) ||
      (patch <= 31)) {
    return 0;
  }

  /* Try to open device */

  fd = open ("/dev/rtc", O_RDWR);
  if (fd < 0) {
    LOG(LOGS_ERR, LOGF_RtcLinux, "Could not open /dev/rtc, %s\n", strerror(errno));
    return 0;
  }

  n_samples = 0;
  n_samples_since_regression = 0;
  n_runs = 0;
  coefs_valid = 0;
  coefs_file_name = file_name;

  measurement_period = LOWEST_MEASUREMENT_PERIOD;

  operating_mode = OM_NORMAL;

  /* Register file handler */
  SCH_AddInputFileHandler(fd, read_from_device, NULL);

  /* Register slew handler */
  LCL_AddParameterChangeHandler(slew_samples, NULL);

  if (CNF_GetLogRtc()) {
    direc = CNF_GetLogDir();
    if (!mkdir_and_parents(direc)) {
      LOG(LOGS_ERR, LOGF_RtcLinux, "Could not create directory %s\n", direc);
      logfile = NULL;
    } else {
      sprintf(filename, "%s/rtc.log", direc);
      logfile = fopen(filename, "a");
      if (!logfile) {
        LOG(LOGS_WARN, LOGF_RtcLinux, "Couldn't open logfile %s for update\n", filename);
      }
    }
  }

  return 1;
}

/* ================================================== */

void
RTC_Linux_Finalise(void)
{
  if (timeout_running) {
    SCH_RemoveTimeout(timeout_id);
    timeout_running = 0;
  }

  /* Remove input file handler */
  if (fd >= 0) {
    SCH_RemoveInputFileHandler(fd);
    close(fd);

    /* Save the RTC data */
    (void) RTC_Linux_WriteParameters();

  }

  if (logfile) {
    fclose(logfile);
  }

}

/* ================================================== */

static void
switch_interrupts(int onoff)
{
  int status;

  if (onoff) {
    status = ioctl(fd, RTC_UIE_ON, 0);
    if (status < 0) {
      LOG(LOGS_ERR, LOGF_RtcLinux, "Could not start measurement : %s\n", strerror(errno));
      return;
    }
  } else {
    status = ioctl(fd, RTC_UIE_OFF, 0);
    if (status < 0) {
      LOG(LOGS_ERR, LOGF_RtcLinux, "Could not stop measurement : %s\n", strerror(errno));
      return;
    }
  }
}    

/* ================================================== */

static void
measurement_timeout(void *any)
{
  timeout_running = 0;
  switch_interrupts(1);
}

/* ================================================== */

static void
set_rtc(time_t new_rtc_time)
{
  struct tm rtc_tm;
  struct rtc_time rtc_raw;
  int status;

  rtc_tm = *gmtime(&new_rtc_time);
  rtc_raw.tm_sec = rtc_tm.tm_sec;
  rtc_raw.tm_min = rtc_tm.tm_min;
  rtc_raw.tm_hour = rtc_tm.tm_hour;
  rtc_raw.tm_mday = rtc_tm.tm_mday;
  rtc_raw.tm_mon = rtc_tm.tm_mon;
  rtc_raw.tm_year = rtc_tm.tm_year;

  status = ioctl(fd, RTC_SET_TIME, &rtc_raw);
  if (status < 0) {
    LOG(LOGS_ERR, LOGF_RtcLinux, "Could not set RTC time\n");
  }

}

/* ================================================== */

static void
handle_initial_trim(void)
{
  FILE *in;
  double ref_offset, rate_ppm, rate;
  int valid;
  time_t ref_time;
  char line[1024];
  long delta_time;
  double rtc_error_now, sys_error_now;

    /* The idea is to accumulate some number of samples at 1 second
       intervals, then do a robust regression fit to this.  This
       should give a good fix on the intercept (=system clock error
       rel to RTC) at a particular time, removing risk of any
       particular sample being an outlier.  We can then look at the
       elapsed interval since the epoch recorded in the RTC file,
       and correct the system time accordingly. */
    
  run_regression(1, &coefs_valid, &coef_ref_time, &coef_seconds_fast, &coef_gain_rate);

  n_samples_since_regression = 0;
  n_samples = 0;
      
  in = fopen(coefs_file_name, "r");
  if (in) {
    if (fgets(line, sizeof(line), in)) {
      if (sscanf(line, "%d%ld%lf%lf", &valid, &ref_time, &ref_offset, &rate_ppm) == 4) {
        if (valid) {
          /* Can process data */
          delta_time = coef_ref_time - ref_time;
          rate = 1.0e-6 * rate_ppm;
          rtc_error_now = ref_offset + rate * (double) delta_time;
          
          /* sys_error_now is positive if the system clock is fast */
          sys_error_now = rtc_error_now - coef_seconds_fast;
          
          LOG(LOGS_INFO, LOGF_RtcLinux, "Elapsed = %ld seconds, sys_error_now=%f\n", delta_time, sys_error_now);
          LCL_ApplyStepOffset(sys_error_now);
        }              
      } else {
        LOG(LOGS_WARN, LOGF_RtcLinux, "Could not parse coefficients line from %s\n", coefs_file_name);
      }
    } else {
      LOG(LOGS_WARN, LOGF_RtcLinux, "Could not read first line from %s\n", coefs_file_name);
    }
    fclose(in);
  } else {
    LOG(LOGS_WARN, LOGF_RtcLinux, "Could not open %s\n", coefs_file_name);
  }
  
  coefs_valid = 0;
  
  (after_init_hook)(after_init_hook_arg);
  
  operating_mode = OM_NORMAL;

  return;
}

/* ================================================== */

static void
handle_relock_after_trim(void)
{
  int valid;
  time_t ref;
  double fast, slope;
  FILE *out;

  run_regression(1, &valid, &ref, &fast, &slope);

  if (valid) {
    out = fopen(coefs_file_name, "w");
    if (out) {
      fprintf(out, "1 %ld %.6f %.3f\n",
              ref, fast, 1.0e6 * saved_coef_gain_rate);
      fclose(out);
    }
  } else {
    LOG(LOGS_WARN, LOGF_RtcLinux, "Could not do regression after trim\n");
  }

  n_samples = 0;
  n_samples_since_regression = 0;
  operating_mode = OM_NORMAL;
  measurement_period = LOWEST_MEASUREMENT_PERIOD;
}

/* ================================================== */

/* Day number of 1 Jan 1970 */
#define MJD_1970 40587

static void
process_reading(time_t rtc_time, struct timeval *system_time)
{
  double rtc_fast;
  int mjd, hour, minute, second, millisecond;

  accumulate_sample(rtc_time, system_time);

  switch (operating_mode) {
    case OM_NORMAL:

      if (n_samples_since_regression >= /* 4 */ 1 ) {
        run_regression(1, &coefs_valid, &coef_ref_time, &coef_seconds_fast, &coef_gain_rate);
        n_samples_since_regression = 0;
      }
      
      break;
    case OM_INITIAL:
      if (n_samples_since_regression >= 8) {
        handle_initial_trim();
      }
      break;
    case OM_AFTERTRIM:
      if (n_samples_since_regression >= 8) {
        handle_relock_after_trim();
      }
      break;
    default:
      assert(0);
      break;
  }  


  if (logfile) {
    mjd = system_time->tv_sec / 86400 + MJD_1970;
    hour = (system_time->tv_sec % 86400) / 3600;
    minute = (system_time->tv_sec % 3600) / 60;
    second = (system_time->tv_sec % 60);
    millisecond = system_time->tv_usec / 1000;

    rtc_fast = (double)(rtc_time - system_time->tv_sec) - 1.0e-6 * (double) system_time->tv_usec;
    
    fprintf(logfile, "%5d %02d:%02d:%02d.%03d  %14.6f %1d  %14.6f  %12.3f  %2d  %2d %4d\n",
            mjd, hour, minute, second, millisecond,
            rtc_fast,
            coefs_valid,
            coef_seconds_fast, coef_gain_rate * 1.0e6, n_samples, n_runs, measurement_period);

    fflush(logfile);
  }    

}

/* ================================================== */

static void
read_from_device(void *any)
{
  int status;
  unsigned long data;
  struct timeval sys_time;
  struct rtc_time rtc_raw;
  struct tm rtc_tm;
  time_t rtc_t;
  double read_err;
  int error = 0;

  status = read(fd, &data, sizeof(data));
  if (status < 0) {
    LOG(LOGS_ERR, LOGF_RtcLinux, "Could not read flags /dev/rtc : %s\n", strerror(errno));
    error = 1;
    goto turn_off_interrupt;
  }    

  if ((data & RTC_UIE) == RTC_UIE) {
    /* Update interrupt detected */
    
    /* Read RTC time, sandwiched between two polls of the system clock
       so we can bound any error. */

    LCL_ReadCookedTime(&sys_time, &read_err);

    status = ioctl(fd, RTC_RD_TIME, &rtc_raw);
    if (status < 0) {
      LOG(LOGS_ERR, LOGF_RtcLinux, "Could not read time from /dev/rtc : %s\n", strerror(errno));
      error = 1;
      goto turn_off_interrupt;
    }

    /* Convert RTC time into a struct timeval */
    rtc_tm.tm_sec = rtc_raw.tm_sec;
    rtc_tm.tm_min = rtc_raw.tm_min;
    rtc_tm.tm_hour = rtc_raw.tm_hour;
    rtc_tm.tm_mday = rtc_raw.tm_mday;
    rtc_tm.tm_mon = rtc_raw.tm_mon;
    rtc_tm.tm_year = rtc_raw.tm_year;
    rtc_tm.tm_isdst = 0;
    
    rtc_t = mktime (&rtc_tm);
    if (rtc_t == (time_t)(-1)) {
      LOG(LOGS_ERR, LOGF_RtcLinux, "Could not convert RTC time to timeval\n");
      error = 1;
      goto turn_off_interrupt;
    }      

    process_reading(rtc_t, &sys_time);

    if (n_samples < 4) {
      measurement_period = LOWEST_MEASUREMENT_PERIOD;
    } else if (n_samples < 6) {
      measurement_period = LOWEST_MEASUREMENT_PERIOD << 1;
    } else if (n_samples < 10) {
      measurement_period = LOWEST_MEASUREMENT_PERIOD << 2;
    } else if (n_samples < 14) {
      measurement_period = LOWEST_MEASUREMENT_PERIOD << 3;
    } else {
      measurement_period = LOWEST_MEASUREMENT_PERIOD << 4;
    }

  }

turn_off_interrupt:

  switch (operating_mode) {
    case OM_INITIAL:
      if (error) {
        LOG(LOGS_WARN, LOGF_RtcLinux, "Could not complete initial step due to errors");
        operating_mode = OM_NORMAL;
        (after_init_hook)(after_init_hook_arg);

        switch_interrupts(0);
    
        timeout_running = 1;
        timeout_id = SCH_AddTimeoutByDelay((double) measurement_period, measurement_timeout, NULL);
      }

      break;

    case OM_AFTERTRIM:
      if (error) {
        LOG(LOGS_WARN, LOGF_RtcLinux, "Could not complete after trim relock due to errors");
        operating_mode = OM_NORMAL;

        switch_interrupts(0);
    
        timeout_running = 1;
        timeout_id = SCH_AddTimeoutByDelay((double) measurement_period, measurement_timeout, NULL);
      }
      
      break;

    case OM_NORMAL:
      switch_interrupts(0);
    
      timeout_running = 1;
      timeout_id = SCH_AddTimeoutByDelay((double) measurement_period, measurement_timeout, NULL);

      break;
    default:
      assert(0);
      break;
  }

}

/* ================================================== */

void
RTC_Linux_TimeInit(void (*after_hook)(void *), void *anything)
{
  after_init_hook = after_hook;
  after_init_hook_arg = anything;

  operating_mode = OM_INITIAL;
  timeout_running = 0;
  switch_interrupts(1);

}

/* ================================================== */

void
RTC_Linux_StartMeasurements(void)
{
  timeout_running = 0;
  measurement_timeout(NULL);
}

/* ================================================== */

int
RTC_Linux_WriteParameters(void)
{
  FILE *out;

  if (fd < 0) {
    return RTC_ST_NODRV;
  }
  
  out = fopen(coefs_file_name, "w");
  if (!out) {
    return RTC_ST_BADFILE;
  }

  if (coefs_valid) {
    /* Gain rate is written out in ppm */
    fprintf(out, "1 %ld %.6f %.3f\n",
            coef_ref_time, coef_seconds_fast, 1.0e6 * coef_gain_rate);


  } else {
    fprintf(out, "0 0 0.0 0.0\n");
  }


  fclose(out);
  return RTC_ST_OK;

}

/* ================================================== */
/* Try to set the system clock from the RTC, in the same manner as
   /sbin/clock -s -u would do.  We're not as picky about OS version
   etc in this case, since we have fewer requirements regarding the
   RTC behaviour than we do for the rest of the module. */

void
RTC_Linux_TimePreInit(void)
{
  int fd, status;
  struct rtc_time rtc_raw;
  struct tm rtc_tm;
  time_t rtc_t;
  struct timeval new_sys_time;

  fd = open("/dev/rtc", O_RDONLY);

  if (fd < 0) {
    return; /* Can't open it, and won't be able to later */
  }

  status = ioctl(fd, RTC_RD_TIME, &rtc_raw);

  if (status >= 0) {
    /* Convert to seconds since 1970 */
    rtc_tm.tm_sec = rtc_raw.tm_sec;
    rtc_tm.tm_min = rtc_raw.tm_min;
    rtc_tm.tm_hour = rtc_raw.tm_hour;
    rtc_tm.tm_mday = rtc_raw.tm_mday;
    rtc_tm.tm_mon = rtc_raw.tm_mon;
    rtc_tm.tm_year = rtc_raw.tm_year;
    rtc_tm.tm_isdst = 0;

#if 0
    LOG(LOGS_INFO, LOGF_RtcLinux, "s=%d m=%d h=%d d=%d m=%d y=%d\n",
        rtc_tm.tm_sec, rtc_tm.tm_min, rtc_tm.tm_hour,
        rtc_tm.tm_mday, rtc_tm.tm_mon, rtc_tm.tm_year);
#endif
    
    rtc_t = mktime (&rtc_tm);
    if (rtc_t != (time_t)(-1)) {

      new_sys_time.tv_sec = rtc_t;
      new_sys_time.tv_usec = 0;

      /* Tough luck if this fails */
      settimeofday(&new_sys_time, NULL);
    }
  }

  close(fd);
}

/* ================================================== */

int
RTC_Linux_GetReport(RPT_RTC_Report *report)
{
  report->ref_time = (unsigned long) coef_ref_time;
  report->n_samples = n_samples;
  report->n_runs = n_runs;
  if (n_samples > 1) {
    report->span_seconds = ((rtc_sec[n_samples-1] - rtc_sec[0]) +
                            (long)(rtc_trim[n_samples-1] - rtc_trim[0]));
  } else {
    report->span_seconds = 0;
  }
  report->rtc_seconds_fast = coef_seconds_fast;
  report->rtc_gain_rate_ppm = 1.0e6 * coef_gain_rate;
  return 1;
}

/* ================================================== */

int
RTC_Linux_Trim(void)
{
  struct timeval now;
  double local_clock_err;


  /* Remember the slope coefficient - we won't be able to determine a
     good one in a few seconds when we determine the new offset! */
  saved_coef_gain_rate = coef_gain_rate;

  if (fabs(coef_seconds_fast) > 1.0) {

    LOG(LOGS_INFO, LOGF_RtcLinux, "Trimming RTC, error = %.3f seconds\n", coef_seconds_fast);

    /* Do processing to set clock.  Let R be the value we set the
       RTC to, then in 500ms the RTC ticks (R+1) (see comments in
       arch/i386/kernel/time.c about the behaviour of the real time
       clock chip).  If S is the system time now, the error at the
       next RTC tick is given by E = (R+1) - (S+0.5).  Ideally we
       want |E| <= 0.5, which implies R <= S <= R+1, i.e. R is just
       the rounded down part of S, i.e. the seconds part. */

    LCL_ReadCookedTime(&now, &local_clock_err);
    
    set_rtc(now.tv_sec);

    /* All old samples will now look bogus under the new
           regime. */
    n_samples = 0;
    operating_mode = OM_AFTERTRIM;

    /* And start rapid sampling, interrupts on now */
    if (timeout_running) {
      SCH_RemoveTimeout(timeout_id);
      timeout_running = 0;
    }
    switch_interrupts(1);
  }
  
}

/* ================================================== */

#endif /* defined LINUX */
